/*====================================================================================

  This sketch demonstrates loading images which have been stored as files in the
  built-in FLASH memory on a NodeMCU (ESP8266 based) rendering them on a TFT screen
  that uses a ILI9341 driver chip with SPI interface.

  The images are stored in the SPI FLASH Filing System (SPIFFS), which effectively
  functions like a tiny "hard drive". This filing system is built into the ESP8266
  Core that can be loaded from the IDE "Boards manager" menu option. This is at
  version 2.3.0 at the time of sketch creation.

  The size of the SPIFFS partition can be set in the IDE as 1Mbyte or 3Mbytes. Either
  will work with this sketch. Typically most sketches easily fit within 1 Mbyte so a
  3 Mbyte SPIFS partition can be used, in which case it can contain ~18 full screen
  320 x 240 raw images (150 Kbytes each).

  The NodeMCU, TFT and sketch works with TFT_ILI9341_ESP library here:
  https://github.com/Bodmer/TFT_ILI9341_ESP

  Configure the SPI speed to 40MHz maximum in the User_Setup file within the
  TFT_ILI9341_ESP library folder. It *may* work at 80MHz, but typically pixels get lost
  or spurious extra pixels are plotted at very high SPI clock rates.

  Images in the 565 format can be created with the nice UTFT library tool from:
  http://www.rinkydinkelectronics.com/library.php?id=51

  The tool can be found is "libraries\UTFT\Tools\" and is called ImageConverter565.exe
  Save the converted image as a .raw file (see option in the tool GUI). It will handle
  BMP and JPEG files.

  The UTFT tool stores the image pixel bytes in the order compatible with this sketch.
  A 320 x 240 x 16 bit image has a file size of ~150k bytes, smaller images scale
  pro-rata in terms of space needed.

  NOTE: The ILI9341_Due library image conversion tool will generate images in a
  different format, the pixel bytes are in a different order :-(  it also includes
  a 54 byte BMP header. This sketch can handle that format but rendering will be
  slower for technical reasons (i.e. 32 bit SPI buffer byte send order).
  https://github.com/marekburiak/ILI9341_due
  
  Place the images inside the sketch folder, in a folder called "Data".  Then upload
  all the files in the folder using the Arduino IDE "ESP8266 Sketch Data Upload" option
  in the "Tools" menu.  This takes some time, but the SPIFFS content is not altered
  when a new sketch is uploaded, so there is no need to upload the same files again!
  Note: If open, you must close the "Serial Monitor" window for it to upload!

  This sketch includes example images in the Data folder, 2 in "UTFT" format, one in
  "Due" format.

  Saving images, uploading and rendering on the TFT screen couldn't be much easier!

  The typical setup for a NodeMCU1.0 (ESP-12 Module) is :

  Display SDO/MISO      to NodeMCU pin D6 <<<<<< This is not used by the library
  Display LED           to NodeMCU pin 3.3V
  Display SCK           to NodeMCU pin D5
  Display SDI/MOSI      to NodeMCU pin D7
  Display DC/RS (or AO) to NodeMCU pin D3
  Display RESET         to NodeMCU pin D4
  Display CS            to NodeMCU pin D8
  Display GND           to NodeMCU pin GND
  Display VCC           to NodeMCU pin 3.3V

  Note: only some versions of the NodeMCU provide the USB 5V on the VIN pin
  If 5V is not available at a pin you can use 3.3V but backlight brightness
  will be lower.

  Created by Bodmer 23rd Mar 2019 - Tested in Arduino IDE 1.8.0 esp8266 Core 2.4.2
  ==================================================================================*/

//====================================================================================
//                                  Libraries
//====================================================================================
//Call up the SPIFFS FLASH filing system
#include <FS.h>
#include <ESP8266WiFi.h>
#include <TFT_eSPI.h> // Hardware-specific library
#include <SPI.h>

TFT_eSPI tft = TFT_eSPI();       // Invoke custom library

//====================================================================================
//                                    Setup
//====================================================================================

void setup()
{
  WiFi.forceSleepBegin(); // Disable wifi, save power
  
  Serial.begin(115200); // Used for messages

  tft.begin();
  tft.setRotation(0);  // 0 & 2 Portrait. 1 & 3 landscape
  tft.fillScreen(TFT_BLACK);

  if (!SPIFFS.begin()) {
    Serial.println("SPIFFS initialisation failed!");
    while (1) yield(); // Stay here twiddling thumbs waiting
  }
  Serial.println("\r\nInitialisation done.");

  listFiles(); // Lists the files so you can see what is in the SPIFFS
  delay(2000);
}

//====================================================================================
//                                    Loop
//====================================================================================
void loop()
{
  // Note the / before the SPIFFS file name must be present, this means the file is in
  // the root directory of the SPIFFS, e.g. "/Tiger.raw" for a file called "Tiger.raw"


  // Plot at x,y = 0,0 an image 240 pixels wide and 320 pixels high.
  drawRaw("/screen1.raw", 0, 0, 240, 320);
  delay(10000); // Hang around for a while

  tft.fillScreen(TFT_BLACK);
  
  drawRaw("/screen2.raw", 0, 0, 240, 320);
  delay(10000);  // Hang around for a while

  uint32_t drawTime = micros(); // Save current time for performance evaluation
  tft.fillScreen(TFT_BLACK);
  drawTime = micros() - drawTime; // Calculate the time it took
  Serial.println("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
  Serial.print("Time to clear screen was : "); Serial.print(drawTime / 1000.0, 1); Serial.println(" ms");
  Serial.print("Average SPI bit rate was : "); Serial.print( (11 * 8 + 320 * 240 * 16.0) / drawTime, 1); Serial.println(" Mbps");
  Serial.println("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
  delay(500);  // Hang around for a while
}
